import 'dart:io';
import 'package:riedev_graph/graph.dart';

typedef String NodeLabelFunc<TNode>(TNode node, int idx);
typedef String EdgeLabelFunc<TEdge>(TEdge edge, int sourceIdx, int targetIdx);

enum NodeShape {
  diamond, box, circle, record, plaintext
}

class NodeStyle<TNode, TEdge extends Edge<TNode>> {
  
  final DotFile<TNode, TEdge> _exporter;
  
  NodeShape _shape;
  String _color;
  String _fillcolor;
  NodeLabelFunc _labelFunc;

  NodeStyle(this._exporter) 
    : _shape = NodeShape.circle;

  NodeStyle setShape(NodeShape shape) {
    _shape = shape;
    return this;
  }

  NodeStyle setColor(String color) {
    _color = color;
    return this;
  }

  NodeStyle setFillColor(String color) {
    _fillcolor = color;
    return this;
  }

  NodeStyle setLabel(NodeLabelFunc labelFunc) {
    _labelFunc = labelFunc;
    return this;
  }

  DotFile<TNode, TEdge> apply() => _exporter;
}

enum EdgeShape {
  dashed, filled
}

class EdgeStyle<TNode, TEdge extends Edge<TNode>> {
  
  final DotFile<TNode, TEdge> _exporter;
  
  EdgeShape _shape;
  String _color;
  EdgeLabelFunc _labelFunc;
  bool _nonArrowHead;
  double _penWidth;
  bool _isPenWidthSet;

  EdgeStyle(this._exporter) 
    : _shape = EdgeShape.filled, _isPenWidthSet = false, _nonArrowHead = false;

  EdgeStyle setShape(EdgeShape shape) {
    _shape = shape;
    return this;
  }

  EdgeStyle setColor(String color) {
    _color = color;
    return this;
  }

  EdgeStyle setLabel(EdgeLabelFunc labelFunc) {
    _labelFunc = labelFunc;
    return this;
  }

  EdgeStyle nonArrowHead() {
    _nonArrowHead = true;
    return this;
  }

  EdgeStyle penWidth(double width) {
    _penWidth = width;
    _isPenWidthSet = true;
    return this;
  }

  DotFile<TNode, TEdge> apply() => _exporter;
}

class NodeStyleBuilder<TNode, TEdge extends Edge<TNode>> extends NodeStyle<TNode, TEdge> {
  
  NodeStyleBuilder(DotFile<TNode, TEdge> exporter) 
    : super(exporter);

  String Build(TNode node, int idx) {
    return "$idx " + _getStyleString(node, idx);
  }

  String _getStyleString(TNode node, int idx) {  
    var style = "[";

    if (_labelFunc != null) style += "label=\"${_labelFunc(node, idx)}\"";
    if (_color != null) style += style.length == 1 ? "color=\"$_color\"" : " color=\"$_color\"";
    if (_fillcolor != null) style += style.length == 1 ? "style=\"filled\" fillcolor=\"$_fillcolor\"" : " style=\"filled\" fillcolor=\"$_fillcolor\"";
    
    style += style.length == 1 ? "shape=\"${_getEnumValue(_shape)}\"" : " shape=\"${_getEnumValue(_shape)}\"";
    
    return style + "];";
  }

  String _getEnumValue(NodeShape shape) {
    return "${shape.toString().substring(shape.toString().indexOf('.')+1)}";
  }
}

class EdgeStyleBuilder<TNode, TEdge extends Edge<TNode>> extends EdgeStyle<TNode, TEdge> {
  
  EdgeStyleBuilder(DotFile<TNode, TEdge> exporter) 
    : super(exporter);

  String Build(TEdge edge, int sourceIdx, int targetIdx) {
    return "$sourceIdx -> $targetIdx " + _getStyleString(edge, sourceIdx, targetIdx);
  }

  String _getStyleString(TEdge edge, int sourceIdx, int targetIdx) {  
    var style = "[";

    if (_labelFunc != null) style += "label=\"${_labelFunc(edge, sourceIdx, targetIdx)}\"";
    if (_color != null) style += style.length == 1 ? "color=\"$_color\"" : " color=\"$_color\"";
        
    style += style.length == 1 ? "style=\"${_getEnumValue(_shape)}\"" : " style=\"${_getEnumValue(_shape)}\"";
    
    if(_nonArrowHead)
      style += style.length == 1 ? "arrowhead=\"none\"" : " arrowhead=\"none\"";

    if(_isPenWidthSet)
      style += style.length == 1 ? "penwidth=\"$_penWidth\"" : " penwidth=\"$_penWidth\"";

    return style + "];";
  }

  String _getEnumValue(EdgeShape shape) {
    return "${shape.toString().substring(shape.toString().indexOf('.')+1)}";
  }
}

class DotFile<TNode, TEdge extends Edge<TNode>> {

  String _filename;
  
  NodeStyleBuilder _nodeStyle;
  final Map<TNode, NodeStyleBuilder> _nodeStyles = Map();

  EdgeStyleBuilder _edgeStyle;
  final Map<TEdge, EdgeStyleBuilder> _edgeStyles = Map();
  
  final Map<Graph<TNode, TEdge>, String> _graphs = Map();

  DotFile() {
    _nodeStyle = NodeStyleBuilder(this);
    _edgeStyle = EdgeStyleBuilder(this);
  }

  void export() async {
    var file = new File(_filename ?? 'graph.dot');
    var sink = file.openWrite();

    for (var graph in _graphs.keys) {
      _writeGraph(sink, graph, _graphs[graph]);
      sink.writeln();
    }

    sink.close();
  }

  DotFile<TNode, TEdge> addGraph(Graph<TNode, TEdge> graph,
      {String graphName = null}) {
    _graphs[graph] = graphName;
    return this;
  }

  DotFile<TNode, TEdge> setFileName(String filename) {
    _filename = filename;
    return this;
  }

  NodeStyle<TNode, TEdge> setNodeStyle({TNode node = null}) {
    var style = new NodeStyleBuilder(this);
    
    if(node == null)
      _nodeStyle = style;
    else _nodeStyles[node] = style;

    return style;
  }

  EdgeStyle<TNode, TEdge> setEdgeStyle({TEdge edge = null}) {
    var style = new EdgeStyleBuilder(this);
    
    if(edge == null)
      _edgeStyle = style;
    else _edgeStyles[edge] = style;

    return style;
  }

  void _writeGraph(IOSink sink, Graph<TNode, TEdge> graph, String graphName) {
    sink.writeln("digraph $graphName {");

    var idx = 0;
    var nodeIndices = Map<TNode, int>();

    for (var node in graph.Nodes) {
      sink.writeln(_getNodeString(node, idx));
      nodeIndices[node] = idx++;
    }

    for (var edge in graph.Edges) {
      sink.writeln(_getEdgeString(edge, nodeIndices[edge.source], nodeIndices[edge.target]));
    }

    sink.writeln("}");
  }

  String _getNodeString(TNode node, int idx) {
    if(_nodeStyles.containsKey(node))
      return _nodeStyles[node].Build(node, idx);
    return _nodeStyle.Build(node, idx);
  }

  String _getEdgeString(TEdge edge, int sourceIdx, int targetIdx) {
    if(_edgeStyles.containsKey(edge))
      return _edgeStyles[edge].Build(edge, sourceIdx, targetIdx);
    return _edgeStyle.Build(edge, sourceIdx, targetIdx);
  }
}